"""
Description
===========

Module: ``geocompy.gsi``

The ``gsi`` package provides wrapper methods for GSI Online commands,
supported by many totalstations (TPS) and digital levels (NA/DNA).


Subpackages
-----------

``geocompy.gsi.dna``
    GSI Online digital level commands.

Submodules
----------

- ``geocompy.gsi.gsidata``
- ``geocompy.gsi.gsitypes``
"""
